#!/usr/bin/perl

my $license = "
    gbk_reformat.pl
    Copyright (C) 2014 Egon A. Ozer

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see [http://www.gnu.org/licenses/].
";

my $version = "0.1.3";

## based on resort_RAST_gbk.pl and gbk_joiner_v0.2.pl

##Changes from v0.1.2 -> 0.1.3
# Changed sort function in subroutine to ignore "<" or ">" in coordinates
# Fixed bug where product lines with the word "source" in them would cause records to be dropped

##Changes from v0.1 -> 0.1.2
# Accept files with all types of line endings (Unix, Mac, PC)

use strict;
use Cwd;
use warnings;

my $usage = "
USAGE:
gbk_reformat.pl [options] <file.gbk> (<file2.gbk> <file3.gbk> ...) > output.gbk

Copyright (C) 2014  Egon A. Ozer
This program comes with ABSOLUTELY NO WARRANTY. This is free software, and you
are welcome to redistribute it under certain conditions;
see LICENSE.txt for details

Sorts records in genbank formatted file(s) to interleave tRNAs, rRNAs, and
other records with CDSs.

If records are missing locus tags, these can be added.  Locus tags will have
the format \"prefix_00000\" where \"prefix\" is profided by the user and
\"00000\" is an ascending number assigned to each record in the output order.
If a prefix is provided and any records in the input file(s) already contain
locus_tag tags, the program will exit with an error.

Multi-sequence genbank files and/or multiple genbank files can be joined into
a single sequence.

Options:
  -l <string>       locus tag prefix
                    (default: no locus tags added)
  -j <no argument>  join input sequences
                    (default: sequences will remain separate)
  -n <integer>      number of N's with which to separate joined sequences.  
                    (default: 200)
  -i <string>       source ID for the joined genbank record. Will be trimmed
                    down to 20 characters if too long.
                    (default: \"organism\")

";

use Getopt::Std;
our ($opt_l, $opt_j, $opt_n, $opt_i);
getopts('l:jn:i:');
die $usage unless (@ARGV >= 1);

my $pref    = $opt_l if $opt_l;
my $ns      = $opt_n ? $opt_n : 200;
my $src_id  = $opt_i ? $opt_i : "organism";

my $in_cds = 0;
my $out_cds = 0;

#create the joining N's string
my $spacer;
for my $i (1 .. $ns){
    $spacer .= "n";
}

#set default spacing parameters ... although it's probably safest and easiest to assume these will be constant between genbank files and records.  I'll have to rethink if I encounter many files where this isn't the case.
my $rec_space = 5;
my $subfeat_space = 21;

#initialize parameters
my $g_rec_count = 0;
my @recs;
my $tot_recs = 0;
my $seq;
my $tot_size = 0;

# read in file(s)
for my $i (0 .. $#ARGV){
    my $gbk = $ARGV[$i];
    open (my $in , "<$gbk") or die "Can't open file $gbk: $!\n";
    
    my $rec_count = 0;
    my $rec;
    my ($g_start, $g_stop) = (0) x 2;
    
    my $start_info; #only need to keep headers if sequences aren't being joined, in which case we'll create a new header for the joined molecule
    my $feat_start;
    my $source_header; #again, will only keep if sequences aren't being joined.
    my $parsing_source_header;
    my $leading_space_count = $rec_space;
    my $subfeat_space_count;
    my $base_count; #this is something that RAST adds. We'll only keep it if it's present and we're not joining sequences.
    my $sequence_string; #will save the sequence portion of the record to output unchanged if we're not joining sequences
    my $parsing_sequence;
    my $has_gene_tags; #keep track of whether the file has "gene" records (most) or doesn't (RAST)
    
    while (my $fline = <$in>){
        $fline =~ s/\R/\012/g; #converts to UNIX-style line endings
        my @lines = split("\n", $fline); #need to split lines by line-ending character in the case of Mac-formatted files which only have CR line terminators, not both CR and LF like DOS
        while (@lines){
            my $line = shift @lines;
            $line .= "\n"; #add unix line ending
            next if $line =~ m/^\s*$/; #skip blank lines;
            if (!$feat_start){ #save the genbank file header information. We'll discard later if we're not joining the sequences
                if ($line =~ m/^FEATURES/){
                    $feat_start = 1;
                    #so we only do this once...  This is assuming every genbank record will have a single FEATURES line, which I think is safe.
                    if ($opt_j and $seq){
                        $seq .= $spacer;
                        $tot_size += $ns;
                    }
                }
                $start_info .= $line;
                next;
            }
            if ($line =~ m/^(\s+)source\s+\d+\.\.\d+\s*$/){
                $leading_space_count = length($1);
                $parsing_source_header = 1;
                $source_header .= $line;
                next;
            }
            if ($line =~ m/^\s{$leading_space_count}(\S+)/ and !$parsing_sequence){ #looks for the first line of a record
                my $rec_type = $1; #determine what type of record it is, i.e. gene, CDS, tRNA, rRNA, etc...
                $has_gene_tags = 1 if ($parsing_source_header and $rec_type eq "gene");  #determine whether the genbank record has gene tags for each feature
                $parsing_source_header = "";
                if ($has_gene_tags){
                    if ($rec_type ne "gene"){ #add this to the current record if the genbank record has gene features and the current feature is not a gene (i.e. is CDS, tRNA, etc...)
                        ###change coordinates here, if joining sequences
                        if ($opt_j){
                            $line =~ m/([<>]*\d+)\.\.([<>]*\d+)/;
                            my ($start, $stop) = ($1, $2);
                            (my $s_gt) = $start =~ m/^(\D)/;
                            (my $e_gt) = $stop =~ m/^(\D)/;
                            $start =~ s/^(\D)//;
                            $stop =~ s/^(\D)//;
                            my ($nstart, $nstop) = ($start + $tot_size, $stop + $tot_size);
                            my ($f_start, $f_stop) = ($start, $stop);
                            if ($s_gt){
                                $f_start = $s_gt . $start;
                                $nstart = $s_gt . $nstart;
                            }
                            if ($e_gt){
                                $f_stop = $e_gt . $stop;
                                $nstop = $e_gt . $nstop;
                            }
                            $line =~ s/$f_start\.\.$f_stop/$nstart\.\.$nstop/;
                        }
                        $rec .= $line;
                        next;
                    }
                }
                if ($rec){
                    push @recs, ([$g_start, $g_stop, $rec]);
                    #print STDERR "***g_start: $g_start g_stop: $g_stop\n$rec\n\n";
                    $rec = "";
                }
                # get gene coordinates
                $line =~ m/([<>]*\d+)\.\.([<>]*\d+)/;
                my ($start, $stop) = ($1, $2);
                (my $s_gt) = $start =~ m/^(\D)/;
                (my $e_gt) = $stop =~ m/^(\D)/;
                $start =~ s/^(\D)//;
                $stop =~ s/^(\D)//;
                ($g_start, $g_stop) = ($start, $stop);
                if ($opt_j){ #change the gene coordinates in the record if we're joining sequences
                    my ($nstart, $nstop) = ($start + $tot_size, $stop + $tot_size);
                    my ($f_start, $f_stop) = ($start, $stop);
                    if ($s_gt){
                        $f_start = $s_gt . $start;
                        $nstart = $s_gt . $nstart;
                    }
                    if ($e_gt){
                        $f_stop = $e_gt . $stop;
                        $nstop = $e_gt . $nstop;
                    }
                    $line =~ s/$f_start\.\.$f_stop/$nstart\.\.$nstop/;
                    ($g_start, $g_stop) = ($nstart, $nstop);
                }
                $rec .= $line;
                next;
            }
            if ($line =~ m/^BASE COUNT/){
                $base_count = $line unless $opt_j;
                next;
            }
            if ($line =~ m/^ORIGIN/){
                $parsing_source_header = "";
                if ($rec){
                    push @recs, ([$g_start, $g_stop, $rec]);
                    #print STDERR "***g_start: $g_start g_stop: $g_stop\n$rec\n\n";
                    $rec = "";
                }
                $parsing_sequence = 1;
                next;
            }
            if ($parsing_source_header){
                $source_header .= $line;
                if (!$subfeat_space_count){
                    $line =~ m/^(\s+)\S/;
                    $subfeat_space_count = length($1);
                }
                next;
            }
            if (!$parsing_source_header and !$parsing_sequence){
                die "ERROR: Genbank file already contains locus tags.\n"if ($line =~ m/\s+\/locus_tag=/ and $pref);
                $rec .= $line;
                next;
            }
            if ($line =~ m/^\/\//){
                $parsing_sequence = "";
                ##output everything if we're not joining sequences
                if (!$opt_j){
                    print "$start_info";
                    print "$source_header";
                    if (@recs){
                        sort_recs (\@recs, $leading_space_count, $subfeat_space_count);
                    }
                    print $base_count if $base_count;
                    print "ORIGIN\n";
                    print $sequence_string if $sequence_string;
                    print "//\n";
                    ($start_info, $source_header, $base_count, $sequence_string) = ("") x 4;
                    @recs = ();
                } 
                $feat_start = "";
                next;
            }
            if ($parsing_sequence){
                if (!$opt_j){
                    $sequence_string .= $line;
                    next;
                }
                chomp $line;
                $line =~ s/\s//g; #take out all spaces
                $line =~ s/\d//g; #take out all digits
                $seq .= $line;
                $tot_size += length($line);
                next;
            }
        }
    }
    close ($in);
}

#output the joined sequences
if ($opt_j){
    my $l_space_string;
    for my $i (1 .. $rec_space){
        $l_space_string .= " ";
    }
    my $s_space_string;
    for my $i (1 .. $subfeat_space){
        $s_space_string .= " ";
    }
    
    print "LOCUS       ";
    my $sid = substr($src_id, 0, 20); #cuts the source id down to no more than 20 characters
    printf ("%-20s", $sid);
    printf ("%11s", "$tot_size bp");
    print "    DNA     linear   UNK\n";
    print "DEFINITION  $sid\n";
    print "ACCESSION   unknown\n";
    print "FEATURES             Location/Qualifiers\n";
    print "     source          1..$tot_size\n";
    print "                     /mol_type=\"genomic DNA\"\n";
    print "                     /organism=\"$sid\"\n";
    if (@recs){
        sort_recs (\@recs, $rec_space, $subfeat_space);
    }
    @recs = ();
    print "ORIGIN\n";
    my $seq_count = 1;
    for (my $i = 0; $i < $tot_size; $i +=  60){
        my $segment = substr($seq, $i, 60);
        printf ("%9s", "$seq_count");
        my $seg_leng = length($segment);
        for (my $j = 0; $j < $seg_leng; $j += 10){
            my $subseg = substr($segment, $j, 10);
            print " $subseg";
        }
        print "\n";
        $seq_count += 60;
    }
    print "//\n";
}

#----------------------------------------------------
sub sort_recs {
    my @array = @{$_[0]};
    my $ls = $_[1];
    my $ss = $_[2];
    @array = sort{($a->[0] =~ /\D*(\d+)/)[0] <=> ($b->[0] =~ /\D*(\d+)/)[0] || ($a->[1] =~ /\D*(\d+)/)[0] <=> ($b->[1] =~ /\D*(\d+)/)[0]} @array; #only sorts on integers in records, ignores "<" or ">"
    my $space_string;
    for my $j (1 .. $ss){
        $space_string .= " ";
    }
    for my $j (0 .. $#array){
        my $outline = $array[$j][2];
        if ($pref){ #if we're adding locus tags, do it here
            $tot_recs++;
            my $recnum = sprintf("%05d", $tot_recs);
            my $lid = "$space_string/locus_tag=\"$pref\_$recnum\"";
            my @lines = split("\n", $outline);
            while (@lines){
                my $oline = shift @lines;
                if ($oline =~ m/^\s{$ls}\S+/){
                    print "$oline\n$lid\n";
                } else {
                    print "$oline\n";
                }
            }
        } else {
            print "$outline";
        }
    }
    return (0);
}